"use strict";
/**
 * Copyright (c) 2021 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Sukolsak Sakshuwong <sukolsak@stanford.edu>
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ObjExporter = void 0;
var tslib_1 = require("tslib");
var mesh_builder_1 = require("../../mol-geo/geometry/mesh/mesh-builder");
var sphere_1 = require("../../mol-geo/geometry/mesh/builder/sphere");
var cylinder_1 = require("../../mol-geo/geometry/mesh/builder/cylinder");
var linear_algebra_1 = require("../../mol-math/linear-algebra");
var mol_util_1 = require("../../mol-util");
var color_1 = require("../../mol-util/color/color");
var float_packing_1 = require("../../mol-util/float-packing");
// avoiding namespace lookup improved performance in Chrome (Aug 2020)
var v3fromArray = linear_algebra_1.Vec3.fromArray;
var v3transformMat4 = linear_algebra_1.Vec3.transformMat4;
var v3transformMat3 = linear_algebra_1.Vec3.transformMat3;
var mat3directionTransform = linear_algebra_1.Mat3.directionTransform;
var ObjExporter = /** @class */ (function () {
    function ObjExporter(filename) {
        this.obj = mol_util_1.StringBuilder.create();
        this.mtl = mol_util_1.StringBuilder.create();
        this.vertexOffset = 0;
        this.materialSet = new Set();
        mol_util_1.StringBuilder.writeSafe(this.obj, "mtllib " + filename + ".mtl\n");
    }
    ObjExporter.getSizeFromTexture = function (tSize, i) {
        var r = tSize.array[i * 3];
        var g = tSize.array[i * 3 + 1];
        var b = tSize.array[i * 3 + 2];
        return float_packing_1.decodeFloatRGB(r, g, b);
    };
    ObjExporter.getSize = function (values, instanceIndex, group) {
        var tSize = values.tSize.ref.value;
        var size = 0;
        switch (values.dSizeType.ref.value) {
            case 'uniform':
                size = values.uSize.ref.value;
                break;
            case 'instance':
                size = ObjExporter.getSizeFromTexture(tSize, instanceIndex) / 100;
                break;
            case 'group':
                size = ObjExporter.getSizeFromTexture(tSize, group) / 100;
                break;
            case 'groupInstance':
                var groupCount = values.uGroupCount.ref.value;
                size = ObjExporter.getSizeFromTexture(tSize, instanceIndex * groupCount + group) / 100;
                break;
        }
        return size * values.uSizeFactor.ref.value;
    };
    ObjExporter.getGroup = function (groups, i) {
        var i4 = i * 4;
        var r = groups[i4];
        var g = groups[i4 + 1];
        var b = groups[i4 + 2];
        if (groups instanceof Float32Array) {
            return float_packing_1.decodeFloatRGB(r * 255 + 0.5, g * 255 + 0.5, b * 255 + 0.5);
        }
        return float_packing_1.decodeFloatRGB(r, g, b);
    };
    ObjExporter.prototype.updateMaterial = function (color, alpha) {
        if (this.currentColor === color && this.currentAlpha === alpha)
            return;
        this.currentColor = color;
        this.currentAlpha = alpha;
        var material = color_1.Color.toHexString(color) + alpha;
        mol_util_1.StringBuilder.writeSafe(this.obj, "usemtl " + material);
        mol_util_1.StringBuilder.newline(this.obj);
        if (!this.materialSet.has(material)) {
            this.materialSet.add(material);
            var _a = color_1.Color.toRgbNormalized(color), r = _a[0], g = _a[1], b = _a[2];
            var mtl = this.mtl;
            mol_util_1.StringBuilder.writeSafe(mtl, "newmtl " + material + "\n");
            mol_util_1.StringBuilder.writeSafe(mtl, 'illum 2\n'); // illumination model
            mol_util_1.StringBuilder.writeSafe(mtl, 'Ns 163\n'); // specular exponent
            mol_util_1.StringBuilder.writeSafe(mtl, 'Ni 0.001\n'); // optical density a.k.a. index of refraction
            mol_util_1.StringBuilder.writeSafe(mtl, 'Ka 0 0 0\n'); // ambient reflectivity
            mol_util_1.StringBuilder.writeSafe(mtl, 'Kd '); // diffuse reflectivity
            mol_util_1.StringBuilder.writeFloat(mtl, r, 1000);
            mol_util_1.StringBuilder.whitespace1(mtl);
            mol_util_1.StringBuilder.writeFloat(mtl, g, 1000);
            mol_util_1.StringBuilder.whitespace1(mtl);
            mol_util_1.StringBuilder.writeFloat(mtl, b, 1000);
            mol_util_1.StringBuilder.newline(mtl);
            mol_util_1.StringBuilder.writeSafe(mtl, 'Ks 0.25 0.25 0.25\n'); // specular reflectivity
            mol_util_1.StringBuilder.writeSafe(mtl, 'd '); // dissolve
            mol_util_1.StringBuilder.writeFloat(mtl, alpha, 1000);
            mol_util_1.StringBuilder.newline(mtl);
        }
    };
    ObjExporter.prototype.addMeshWithColors = function (vertices, normals, indices, groups, vertexCount, drawCount, values, instanceIndex, geoTexture, ctx) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var obj, t, n, tmpV, stride, colorType, tColor, uAlpha, aTransform, currentProgress, i, i, i, color, group, groupCount, group, v1, v2, v3;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        obj = this.obj;
                        t = linear_algebra_1.Mat4();
                        n = linear_algebra_1.Mat3();
                        tmpV = linear_algebra_1.Vec3();
                        stride = geoTexture ? 4 : 3;
                        colorType = values.dColorType.ref.value;
                        tColor = values.tColor.ref.value.array;
                        uAlpha = values.uAlpha.ref.value;
                        aTransform = values.aTransform.ref.value;
                        linear_algebra_1.Mat4.fromArray(t, aTransform, instanceIndex * 16);
                        mat3directionTransform(n, t);
                        currentProgress = (vertexCount * 2 + drawCount) * instanceIndex;
                        return [4 /*yield*/, ctx.update({ isIndeterminate: false, current: currentProgress, max: (vertexCount * 2 + drawCount) * values.uInstanceCount.ref.value })];
                    case 1:
                        _a.sent();
                        i = 0;
                        _a.label = 2;
                    case 2:
                        if (!(i < vertexCount)) return [3 /*break*/, 6];
                        if (!(i % 1000 === 0 && ctx.shouldUpdate)) return [3 /*break*/, 4];
                        return [4 /*yield*/, ctx.update({ current: currentProgress + i })];
                    case 3:
                        _a.sent();
                        _a.label = 4;
                    case 4:
                        v3transformMat4(tmpV, v3fromArray(tmpV, vertices, i * stride), t);
                        mol_util_1.StringBuilder.writeSafe(obj, 'v ');
                        mol_util_1.StringBuilder.writeFloat(obj, tmpV[0], 1000);
                        mol_util_1.StringBuilder.whitespace1(obj);
                        mol_util_1.StringBuilder.writeFloat(obj, tmpV[1], 1000);
                        mol_util_1.StringBuilder.whitespace1(obj);
                        mol_util_1.StringBuilder.writeFloat(obj, tmpV[2], 1000);
                        mol_util_1.StringBuilder.newline(obj);
                        _a.label = 5;
                    case 5:
                        ++i;
                        return [3 /*break*/, 2];
                    case 6:
                        i = 0;
                        _a.label = 7;
                    case 7:
                        if (!(i < vertexCount)) return [3 /*break*/, 11];
                        if (!(i % 1000 === 0 && ctx.shouldUpdate)) return [3 /*break*/, 9];
                        return [4 /*yield*/, ctx.update({ current: currentProgress + vertexCount + i })];
                    case 8:
                        _a.sent();
                        _a.label = 9;
                    case 9:
                        v3transformMat3(tmpV, v3fromArray(tmpV, normals, i * stride), n);
                        mol_util_1.StringBuilder.writeSafe(obj, 'vn ');
                        mol_util_1.StringBuilder.writeFloat(obj, tmpV[0], 100);
                        mol_util_1.StringBuilder.whitespace1(obj);
                        mol_util_1.StringBuilder.writeFloat(obj, tmpV[1], 100);
                        mol_util_1.StringBuilder.whitespace1(obj);
                        mol_util_1.StringBuilder.writeFloat(obj, tmpV[2], 100);
                        mol_util_1.StringBuilder.newline(obj);
                        _a.label = 10;
                    case 10:
                        ++i;
                        return [3 /*break*/, 7];
                    case 11:
                        i = 0;
                        _a.label = 12;
                    case 12:
                        if (!(i < drawCount)) return [3 /*break*/, 16];
                        if (!(i % 3000 === 0 && ctx.shouldUpdate)) return [3 /*break*/, 14];
                        return [4 /*yield*/, ctx.update({ current: currentProgress + vertexCount * 2 + i })];
                    case 13:
                        _a.sent();
                        _a.label = 14;
                    case 14:
                        color = void 0;
                        switch (colorType) {
                            case 'uniform':
                                color = color_1.Color.fromNormalizedArray(values.uColor.ref.value, 0);
                                break;
                            case 'instance':
                                color = color_1.Color.fromArray(tColor, instanceIndex * 3);
                                break;
                            case 'group': {
                                group = geoTexture ? ObjExporter.getGroup(groups, i) : groups[indices[i]];
                                color = color_1.Color.fromArray(tColor, group * 3);
                                break;
                            }
                            case 'groupInstance': {
                                groupCount = values.uGroupCount.ref.value;
                                group = geoTexture ? ObjExporter.getGroup(groups, i) : groups[indices[i]];
                                color = color_1.Color.fromArray(tColor, (instanceIndex * groupCount + group) * 3);
                                break;
                            }
                            case 'vertex':
                                color = color_1.Color.fromArray(tColor, i * 3);
                                break;
                            case 'vertexInstance':
                                color = color_1.Color.fromArray(tColor, (instanceIndex * drawCount + i) * 3);
                                break;
                            default: throw new Error('Unsupported color type.');
                        }
                        this.updateMaterial(color, uAlpha);
                        v1 = this.vertexOffset + (geoTexture ? i : indices[i]) + 1;
                        v2 = this.vertexOffset + (geoTexture ? i + 1 : indices[i + 1]) + 1;
                        v3 = this.vertexOffset + (geoTexture ? i + 2 : indices[i + 2]) + 1;
                        mol_util_1.StringBuilder.writeSafe(obj, 'f ');
                        mol_util_1.StringBuilder.writeInteger(obj, v1);
                        mol_util_1.StringBuilder.writeSafe(obj, '//');
                        mol_util_1.StringBuilder.writeIntegerAndSpace(obj, v1);
                        mol_util_1.StringBuilder.writeInteger(obj, v2);
                        mol_util_1.StringBuilder.writeSafe(obj, '//');
                        mol_util_1.StringBuilder.writeIntegerAndSpace(obj, v2);
                        mol_util_1.StringBuilder.writeInteger(obj, v3);
                        mol_util_1.StringBuilder.writeSafe(obj, '//');
                        mol_util_1.StringBuilder.writeInteger(obj, v3);
                        mol_util_1.StringBuilder.newline(obj);
                        _a.label = 15;
                    case 15:
                        i += 3;
                        return [3 /*break*/, 12];
                    case 16:
                        this.vertexOffset += vertexCount;
                        return [2 /*return*/];
                }
            });
        });
    };
    ObjExporter.prototype.addMesh = function (values, ctx) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var aPosition, aNormal, elements, aGroup, instanceCount, vertexCount, drawCount, instanceIndex;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        aPosition = values.aPosition.ref.value;
                        aNormal = values.aNormal.ref.value;
                        elements = values.elements.ref.value;
                        aGroup = values.aGroup.ref.value;
                        instanceCount = values.instanceCount.ref.value;
                        vertexCount = values.uVertexCount.ref.value;
                        drawCount = values.drawCount.ref.value;
                        instanceIndex = 0;
                        _a.label = 1;
                    case 1:
                        if (!(instanceIndex < instanceCount)) return [3 /*break*/, 4];
                        return [4 /*yield*/, this.addMeshWithColors(aPosition, aNormal, elements, aGroup, vertexCount, drawCount, values, instanceIndex, false, ctx)];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        ++instanceIndex;
                        return [3 /*break*/, 1];
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    ObjExporter.prototype.addLines = function (values, ctx) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/];
            });
        });
    };
    ObjExporter.prototype.addPoints = function (values, ctx) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/];
            });
        });
    };
    ObjExporter.prototype.addSpheres = function (values, ctx) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var center, aPosition, aGroup, instanceCount, vertexCount, instanceIndex, state, i, group, radius, mesh, vertices, normals, indices, groups;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        center = linear_algebra_1.Vec3();
                        aPosition = values.aPosition.ref.value;
                        aGroup = values.aGroup.ref.value;
                        instanceCount = values.instanceCount.ref.value;
                        vertexCount = values.uVertexCount.ref.value;
                        instanceIndex = 0;
                        _a.label = 1;
                    case 1:
                        if (!(instanceIndex < instanceCount)) return [3 /*break*/, 4];
                        state = mesh_builder_1.MeshBuilder.createState(512, 256);
                        for (i = 0; i < vertexCount; i += 4) {
                            v3fromArray(center, aPosition, i * 3);
                            group = aGroup[i];
                            radius = ObjExporter.getSize(values, instanceIndex, group);
                            state.currentGroup = group;
                            sphere_1.addSphere(state, center, radius, 2);
                        }
                        mesh = mesh_builder_1.MeshBuilder.getMesh(state);
                        vertices = mesh.vertexBuffer.ref.value;
                        normals = mesh.normalBuffer.ref.value;
                        indices = mesh.indexBuffer.ref.value;
                        groups = mesh.groupBuffer.ref.value;
                        return [4 /*yield*/, this.addMeshWithColors(vertices, normals, indices, groups, vertices.length / 3, indices.length, values, instanceIndex, false, ctx)];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        ++instanceIndex;
                        return [3 /*break*/, 1];
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    ObjExporter.prototype.addCylinders = function (values, ctx) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var start, end, aStart, aEnd, aScale, aCap, aGroup, instanceCount, vertexCount, instanceIndex, state, i, group, radius, cap, topCap, bottomCap, cylinderProps, mesh, vertices, normals, indices, groups;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        start = linear_algebra_1.Vec3();
                        end = linear_algebra_1.Vec3();
                        aStart = values.aStart.ref.value;
                        aEnd = values.aEnd.ref.value;
                        aScale = values.aScale.ref.value;
                        aCap = values.aCap.ref.value;
                        aGroup = values.aGroup.ref.value;
                        instanceCount = values.instanceCount.ref.value;
                        vertexCount = values.uVertexCount.ref.value;
                        instanceIndex = 0;
                        _a.label = 1;
                    case 1:
                        if (!(instanceIndex < instanceCount)) return [3 /*break*/, 4];
                        state = mesh_builder_1.MeshBuilder.createState(512, 256);
                        for (i = 0; i < vertexCount; i += 6) {
                            v3fromArray(start, aStart, i * 3);
                            v3fromArray(end, aEnd, i * 3);
                            group = aGroup[i];
                            radius = ObjExporter.getSize(values, instanceIndex, group) * aScale[i];
                            cap = aCap[i];
                            topCap = cap === 1 || cap === 3;
                            bottomCap = cap >= 2;
                            cylinderProps = { radiusTop: radius, radiusBottom: radius, topCap: topCap, bottomCap: bottomCap, radialSegments: 32 };
                            state.currentGroup = aGroup[i];
                            cylinder_1.addCylinder(state, start, end, 1, cylinderProps);
                        }
                        mesh = mesh_builder_1.MeshBuilder.getMesh(state);
                        vertices = mesh.vertexBuffer.ref.value;
                        normals = mesh.normalBuffer.ref.value;
                        indices = mesh.indexBuffer.ref.value;
                        groups = mesh.groupBuffer.ref.value;
                        return [4 /*yield*/, this.addMeshWithColors(vertices, normals, indices, groups, vertices.length / 3, indices.length, values, instanceIndex, false, ctx)];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        ++instanceIndex;
                        return [3 /*break*/, 1];
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    ObjExporter.prototype.addTextureMesh = function (values, webgl, ctx) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var GeoExportName, framebuffer, _a, width, height, vertices, normals, groups, vertexCount, instanceCount, drawCount, instanceIndex;
            return tslib_1.__generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        GeoExportName = 'geo-export';
                        if (!webgl.namedFramebuffers[GeoExportName]) {
                            webgl.namedFramebuffers[GeoExportName] = webgl.resources.framebuffer();
                        }
                        framebuffer = webgl.namedFramebuffers[GeoExportName];
                        _a = values.uGeoTexDim.ref.value, width = _a[0], height = _a[1];
                        vertices = new Float32Array(width * height * 4);
                        normals = new Float32Array(width * height * 4);
                        groups = webgl.isWebGL2 ? new Uint8Array(width * height * 4) : new Float32Array(width * height * 4);
                        framebuffer.bind();
                        values.tPosition.ref.value.attachFramebuffer(framebuffer, 0);
                        webgl.readPixels(0, 0, width, height, vertices);
                        values.tNormal.ref.value.attachFramebuffer(framebuffer, 0);
                        webgl.readPixels(0, 0, width, height, normals);
                        values.tGroup.ref.value.attachFramebuffer(framebuffer, 0);
                        webgl.readPixels(0, 0, width, height, groups);
                        vertexCount = values.uVertexCount.ref.value;
                        instanceCount = values.instanceCount.ref.value;
                        drawCount = values.drawCount.ref.value;
                        instanceIndex = 0;
                        _b.label = 1;
                    case 1:
                        if (!(instanceIndex < instanceCount)) return [3 /*break*/, 4];
                        return [4 /*yield*/, this.addMeshWithColors(vertices, normals, undefined, groups, vertexCount, drawCount, values, instanceIndex, true, ctx)];
                    case 2:
                        _b.sent();
                        _b.label = 3;
                    case 3:
                        ++instanceIndex;
                        return [3 /*break*/, 1];
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    ObjExporter.prototype.add = function (renderObject, webgl, ctx) {
        if (!renderObject.state.visible)
            return;
        switch (renderObject.type) {
            case 'mesh':
                return this.addMesh(renderObject.values, ctx);
            case 'lines':
                return this.addLines(renderObject.values, ctx);
            case 'points':
                return this.addPoints(renderObject.values, ctx);
            case 'spheres':
                return this.addSpheres(renderObject.values, ctx);
            case 'cylinders':
                return this.addCylinders(renderObject.values, ctx);
            case 'texture-mesh':
                return this.addTextureMesh(renderObject.values, webgl, ctx);
        }
    };
    ObjExporter.prototype.getData = function () {
        return {
            obj: mol_util_1.StringBuilder.getString(this.obj),
            mtl: mol_util_1.StringBuilder.getString(this.mtl)
        };
    };
    return ObjExporter;
}());
exports.ObjExporter = ObjExporter;
//# sourceMappingURL=export.js.map